import logging
from typing import Optional

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

GREEN = '\x1b[32m'
GREY = '\x1b[38;21m'
BLUE = '\x1b[36m'
YELLOW = '\x1b[38;5;226m'
RED = '\x1b[31m'
BOLD_RED = '\x1b[31;1m'
RESET = '\x1b[0m'

class CustomFormatter(logging.Formatter):

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.formats = {
            logging.DEBUG: "[%(levelname)s] " + self.fmt + RESET,
            logging.INFO: BLUE + "[%(levelname)s] " + RESET + self.fmt + RESET,
            logging.WARNING: YELLOW + "[%(levelname)s] " + RESET + self.fmt + RESET,
            logging.ERROR: RED + "[%(levelname)s] " + RESET + self.fmt + RESET,
            logging.CRITICAL: BOLD_RED + "[%(levelname)s] "+ RESET + self.fmt + RESET
        }

    def format(self, record):
        log_fmt = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

def getLogger(name: str, level: int, fmt: Optional[str] = None):
    if fmt is None:
        fmt = '%(message)s'

    logger = logging.getLogger(name)
    logger.setLevel(level)
    stdout_handler = logging.StreamHandler()
    stdout_handler.setLevel(level)
    stdout_handler.setFormatter(CustomFormatter(fmt))
    logger.addHandler(stdout_handler)
    return logger



from .menu import cng as config
from . import menu as _menu
from .decorators import title, ignore
from .utils import clear_screen, input_splitter, set_default_frontend, set_default_once
from .menu import menu, build_menu, Menu
from .casehandlers import _TreeHandler, _handle_casefunc
from .exceptions import MenuQuit, MenuError



class MenuError(Exception):
    """
    Custom exception for console related stuff since I dont want to catch too many exceptions
    from Python.
    """


class MenuQuit(Exception):
    """
    For exiting all console instances
    """



"""
This module primarily conatins a the construct_funcmap
function. The function constructs the dictionary that contains
the console cases. The dictionary keys consists of a range of integers
and the values are the cases (which are functions).
"""

from inspect import unwrap
from types import FunctionType
from typing import Callable, Dict, Iterable, Optional, Tuple

from meny.config import _CASE_TITLE, _DICT_KEY


def _get_case_name(func: FunctionType) -> str:
    """
    First unwraps function (if wrapped in decorators). Assumes that wrapped functions has the
    __wrapped__ attribute (which will be handled by using functools.wraps). Then returns case
    title if set, else just function name
    """
    funcvars = vars(unwrap(func))
    return funcvars.get(_CASE_TITLE, False) or funcvars.get(_DICT_KEY, False) or func.__name__


def construct_funcmap(
    funcs: Iterable[FunctionType], decorator: Optional[FunctionType] = None
) -> Dict[str, Tuple[str, FunctionType]]:
    """
    Parameters
    ------------
    funcs: Iterable[FunctionType]
    decorator: optional, a decorator to decorate all case functions

    Returns
    --------
    Returns dictionary to be used in console interface

    Keys of dictionary are enumeration 1, 2, 3 ... for interface.
    Each item is a tuple with first element as name of case,
    second element is the function itself:
    ('Scrape OSEBX', function object)
    """
    if not isinstance(funcs, Iterable):
        raise TypeError(f"Unsupported type for functions: got {type(funcs)}")

    if decorator is None:
        decorator = lambda f: f

    return {str(i): (_get_case_name(func), decorator(func)) for i, func in enumerate(funcs, start=1)}


if __name__ == "__main__":
    import subprocess

    subprocess.call(["python3", "example/cases.py"])



# Wait time for invalid input message
MSG_WAIT_TIME = 0.3
DEFAULT_ON_INTERRUPT = "raise"
DEFAULT_ON_BLANK = "return"
DEFAULT_FRONTEND = "auto"
DEFAULT_ONCE = False
DEFAULT_RETURN_MODE = "flat"
DEFAULT_REMEMBER = True
DEFAULT_CLEAR = False
_CASE_TITLE = "__meny_title__"
_CASE_IGNORE = "__meny_ignore__"
_DICT_KEY = "__meny_key_from_input_dict__"
_ROOT = "__meny_root__"



import argparse
import os
import sys
from pathlib import Path
from typing import Union
from .menu import menu
from .menylogger import getLogger, INFO
from .utils import _get_module_cases
import importlib.util
import importlib.machinery
import traceback
import types

logger = getLogger("meny.cli", INFO)


def resolve_path(file: str) -> Path:
    path = Path(file)
    if not path.exists():
        raise FileNotFoundError(path)
    else:
        return path.resolve()


def load_module_from_path(path: Path):
    sys.path.append(str(Path(path).parent))
    loader = importlib.machinery.SourceFileLoader(f'__meny_module_{path.stem}', str(path))
    spec = importlib.util.spec_from_loader(loader.name, loader)
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)
    sys.path.pop()
    return module


def cli():
    parser = argparse.ArgumentParser(
        prog="meny", description="Start a meny on a specified Python file"
    )

    parser.add_argument("pythonfile", type=str, nargs=1, help="a python file to start a menu on")
    parser.add_argument(
        "-r",
        "--repeat",
        help="dont exit meny after running a case",
        action="store_true"
    )

    args = parser.parse_args()

    file = args.pythonfile[0]
    try:
        filepath = resolve_path(file)
    except FileNotFoundError as e:
        logger.error(f"Could not find \x1b[33m{file}\x1b[0m")
        sys.exit(1)

    try:
        module = load_module_from_path(filepath)
    except Exception as e:
        logger.error("".join(traceback.TracebackException.from_exception(e).format()))
        logger.error(f"Something went wrong when attempting to import \x1b[33m{filepath}\x1b[0m")
        logger.error(f"Received error: {e}")
        logger.error(f"Please ensure that \x1b[33m{filepath}\x1b[0m contains valid Python code")
        sys.exit(1)

    cases = _get_module_cases(module)
    if len(cases) == 0:
        logger.info(f"There are no defined functions in \x1b[33m{filepath}\x1b[0m")
        sys.exit(1)

    returnDict = menu(cases, f"Functions in {file}", once=not args.repeat, return_mode='flat')
    values = list(returnDict.values())
    if len(values) == 1:
        print(values[0])
    elif len(values) > 0:
        print(returnDict)
    else:
        sys.exit(1)

    sys.exit(0)



"""
Common stuff for console stuff
"""
import os
import re
from meny import config as cng
from inspect import getmodule, isfunction
from types import FunctionType, ModuleType
from typing import Any, Container, Dict, List
from meny import strings

# *Nix uses clear, windows uses cls
_CLEAR_COMMAND = "cls" if os.name == "nt" else "clear"

RE_ANSI = re.compile(r"\x1b\[[;\d]*[A-Za-z]")  # Taken from tqdm source code, matches escape codes

RE_INPUT = re.compile("[\w.-]+|\[.*?\]|\{.*?\}|\(.*?\)|\".*?\"|'.*?'")


def _assert_supported(arg: Any, paramname: str, supported: Container):
    """
    Assert if argument is supported by checking if 'arg' is in 'supported'

    Example
    --------
    >>> _assert_supported("cat", "animal", ("dog", "rabbit"))
    AssertionError: Got unsupported argument for parameter "animal". Available options are: ('dog', 'rabbit')
    """
    assert arg in supported, (
        f'Got unsupported argument "'
        + strings.YELLOW
        + str(arg)
        + strings.END
        + '" for parameter "'
        + strings.YELLOW
        + paramname
        + strings.END
        + f'".\nAvailable options are: {supported}'
    )


def set_default_frontend(frontend: str):
    """
    Default value for meny.menu(frontend=...)
    Options: (simple, fancy, auto)
    """
    _assert_supported(frontend, "frontend", ("simple", "fancy", "auto"))
    cng.DEFAULT_FRONTEND = frontend


def set_default_once(once: bool):
    """
    Default value for
    meny.menu(once=...)
    """
    _assert_supported(type(once), "once", (bool,))
    cng.DEFAULT_ONCE = once


def set_default_remember(remember: bool):
    """Toggle for remembering last input"""
    _assert_supported(type(remember), "remember", (bool,))
    cng.DEFAULT_REMEMBER = remember


def set_default_clear(remember: bool):
    """Toggle for clearing console before showing menu"""
    _assert_supported(type(remember), "remember", (bool,))
    cng.DEFAULT_CLEAR = remember


def clear_screen() -> None:
    """OS independent terminal clear"""
    os.system(_CLEAR_COMMAND)


def _extract_and_preprocess_functions(dict_: Dict[str, FunctionType]) -> List[FunctionType]:
    """
    Parameters
    -------------
    dict_: Dict[str, FunctionType])

    Extract functions from dictionary. Will also add the dictionary key to the function vars
    (i.e. the __dict__ attribute)

    Python parses top down, and inserts the functions in a dictionary. CPython's dict implementation
    in Python 3.6 iterates through dict items in insertion order. As of Python 3.7 said behavior
    become a standard for Python. Thus the functions should come in definition order.

    See:
    https://softwaremaniacs.org/blog/2020/02/05/dicts-ordered/
    """
    funcs = []
    for key, val in dict_.items():
        if isfunction(val):
            vars(val)[cng._DICT_KEY] = key
            funcs.append(val)
    return funcs


def input_splitter(argstring: str) -> List[str]:
    """
    Split string
    """
    return RE_INPUT.findall(argstring)


def _get_module_cases(module: ModuleType) -> List[FunctionType]:
    """Get all functions defined in module"""
    isUserDefinedFunction = lambda f: isfunction(f)
    return [func for func in vars(module).values() if isUserDefinedFunction(func)]


if __name__ == "__main__":
    import subprocess

    subprocess.call(["python3", "example/cases.py"])



from abc import abstractclassmethod
from typing import Any, Dict, Iterable, List, Optional
import meny
from meny.exceptions import MenuError
from ast import literal_eval
from inspect import getfullargspec, unwrap
from types import FunctionType
from meny.infos import _error_info_case
from meny.exceptions import MenuError


def _handle_args(func: FunctionType, args: Iterable[str]) -> List:
    """
    Handles list of strings that are the arguments using ast.literal_eval.

    E.g. return is [1, "cat", 2.0, False]
                   int  str   float  bool
    """
    # Unwrap in case the function is wrapped
    func = unwrap(func)
    argsspec = getfullargspec(func)
    params = argsspec.args

    if len(args) > len(params):
        raise MenuError(f"Got too many arguments, should be {len(params)}, but got {len(args)}")

    typed_arglist = [None] * len(args)
    try:
        for i, arg in enumerate(args):
            typed_arglist[i] = literal_eval(arg)
    except (ValueError, SyntaxError) as e:
        raise MenuError(
            f"Got arguments: {args}\n" f"But could not evaluate argument at position {i}:\n\t {arg}"
        ) from e
    return typed_arglist


def _handle_casefunc(casefunc: FunctionType, args: List[str], menu: meny.Menu) -> Any:
    program_args = menu.case_args.get(casefunc, ())
    program_kwargs = menu.case_kwargs.get(casefunc, {})
    if program_args or program_kwargs:  # If programmatic arguments
        if args:
            raise MenuError(
                "This function takes arguments progammatically" " and should not be given any arguments"
            )
        return casefunc(*program_args, **program_kwargs)
    elif args:
        # Raises TypeError if wrong number of arguments
        return casefunc(*_handle_args(casefunc, args))
    else:
        # Will raise TypeError if casefunc() actually requires arguments
        return casefunc()


class _CaseHandler:
    @classmethod
    def __call__(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]) -> None:
        # TODO: Should I catch TypeError in the handlers? What if actual TypeError occurs?
        #       Maybe should catch everything and just display it in big red text? Contemplate!
        try:
            cls.onCall(menu, casefunc, args)
        except (TypeError, MenuError) as e:
            _error_info_case(e, casefunc)
        finally:
            cls.afterCallReturn(menu, casefunc, args)

    @abstractclassmethod
    def onCall(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]) -> None:
        """
        Responsibility:
            Call casefunc
            Do whatever else to enforce handler behavior (related to unittests)
        """

    @abstractclassmethod
    def afterCallReturn(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]) -> None:
        """
        Responsibility:
            Hook for clean up after calling method
        """


class _TreeHandler(_CaseHandler):
    _stack: List[dict] = []

    @classmethod
    def onCall(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]):
        if len(cls._stack) == 0:
            cls._stack.append({})
        this_scope = cls._stack[-1]  # Get scope of current cls
        next_scope = this_scope.get(casefunc.__name__, {})  # Create / get next scope
        this_scope[casefunc.__name__] = next_scope  # Insert next scope into old scope
        cls._stack.append(next_scope)
        next_scope["return"] = _handle_casefunc(casefunc, args, menu)

    @classmethod
    def afterCallReturn(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]):
        cls._stack.pop()
        type(menu)._return = cls._stack[-1].copy()  # Set current return scope to previous


class _FlatHandler(_CaseHandler):
    _return: dict = {}

    @classmethod
    def onCall(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]):
        cls._return[casefunc.__name__] = _handle_casefunc(casefunc, args, menu)

    @classmethod
    def afterCallReturn(cls, menu: meny.Menu, casefunc: FunctionType, args: List[str]):
        type(menu)._return = cls._return



import meny
import meny.strings as strings
from typing import Dict, Tuple, Callable


def print_funcmap(funcmap: Dict[str, Tuple[str, Callable]]) -> None:
    """
    Prints a func_map dictionary

    Items should be tuples with first elements as descriptions
    and second elements as function objects
    """
    for key, tup in funcmap.items():
        print(key + ".", tup[0])


def logo_title(title: str) -> None:
    """Prints logo title"""
    print("{:-^40s}".format(title))


def show_cases(funcmap: dict, title=strings.LOGO_TITLE) -> None:
    """Prints function map prettily with a given title"""
    logo_title(title)
    print_funcmap(funcmap)


def interface(cli: meny.Menu):
    print("\x1b[s", end="")  # Save current position
    show_cases(cli.funcmap, cli.title)
    retval = input(f"{strings.ENTER_PROMPT}: ")
    print("\x1b[u\x1b[J", end="")  # Restore saved position
    return retval



from meny import strings
from types import FunctionType
from meny.funcmap import _get_case_name
from inspect import signature
from meny.utils import (
    RE_ANSI,
)


def _error_info_case(error: Exception, func: FunctionType) -> None:
    """
    Used to handle error for cases

    Error: E.g. ValueError

    func: case function
    """
    selected_case_str = (
        f'Selected case: "{strings.YELLOW+_get_case_name(func)+strings.END}"\n'
        f'Case function: "{strings.YELLOW+str(func)+strings.END}"\n'
        f"Function signature: {signature(func)}"
    )
    lenerror = max(map(len, str(error).split("\n")))
    lenerror = max(lenerror, max(map(len, RE_ANSI.sub("", selected_case_str).split("\n"))))
    print(strings.BOLD + strings.RED + f"{' ERROR ':#^{lenerror}}" + strings.END)
    print(selected_case_str)
    print(f'{f" Error message ":=^{lenerror}}')
    print(error)
    print(f'{f"":=^{lenerror}}')
    print()
    print(strings.INPUT_WAIT_PROMPT_MSG)
    input()


def _error_info_parse(error: Exception):
    lenerror = max(map(len, str(error).split("\n")))
    print(strings.BOLD + strings.RED + f"{' ARGUMENT PARSE ERROR ':#^{lenerror}}" + strings.END)
    print(f'{f" Error message ":=^{lenerror}}')
    print(error)
    print(f'{f"":=^{lenerror}}')
    print()
    print(strings.INPUT_WAIT_PROMPT_MSG)
    input()


def print_help(*args, **kwargs) -> None:
    print(
        """
        To exit or return from menu interface (even if you are in a nested menu): Enter q

        To return to parent menu: Enter blank (press enter without giving input)
                                  or enter '..'. If you are in main menu, this 
                                  will exit the menu as well. 

        Enter 'r' to restart the Python program. This is usefull for debugging purposes to 
        easily get code changes.

        Press enter to exit help screen
        """
    )
    input()



"""
Contains the command line interface (CLI) class, along its factory function:
menu()
"""
from time import sleep
from types import FunctionType, ModuleType
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from meny import config as cng
from meny import strings
from meny.funcmap import construct_funcmap
from meny.utils import (
    _assert_supported,
    _extract_and_preprocess_functions,
    _get_module_cases,
    input_splitter,
    clear_screen,
)
from meny.infos import _error_info_parse, print_help
from meny.exceptions import MenuQuit


def raise_interrupt(*args, **kwargs) -> None:
    """
    Raises keyboard interrupt
    """
    raise KeyboardInterrupt


def _quit():
    raise MenuQuit


def _menu_simple(instance) -> str:
    # Import here to fix circular imports
    from meny import simple_interface

    return simple_interface.interface(instance)


def _menu_curses(instance) -> str:
    # Import here to fix circular imports
    try:
        from meny import curses_interface
    except ImportError as e:
        raise ImportError(
            f"Got error :\n\t{e}\n"
            "This is probably caused by inability to import the 'curses' module.\n"
            "The curses module should be a built-in for Unix installations.\n"
            "Windows does not have 'curses' by default, suggested fix:\n\t"
            f"pip install windows-curses\n"
            "windows-curses adds support for the standard Python curses module on Windows."
        ) from e

    return curses_interface.interface(instance)


def _restart() -> None:
    """
    Restart application
    """
    import sys
    import os

    os.execv(sys.executable, [sys.executable] + sys.argv)


class Menu:
    """
    Command Line Interface class
    """

    _return: Optional[dict] = None
    _depth: int = 0
    _return_mode: Optional[str] = None

    def __init__(
        self,
        cases: Iterable[FunctionType],
        title: str,
        *,
        case_args: Optional[Dict[FunctionType, tuple]] = None,
        case_kwargs: Optional[Dict[FunctionType, dict]] = None,
        decorator: Optional[FunctionType] = None,
        frontend: str,
        on_blank: str,
        on_kbinterrupt: str,
        once: bool,
        return_mode: str,
    ):
        """
        Input
        -----
        cases: Iterable of case functions

        title: String to print on top

        on_blank: what to do on empty string (press enter without any input)

        See docstring of menu function for more info
        """
        cases = list(cases)
        assert cases, "Given argument for cases is falsey, is it empty?"
        _assert_supported(on_kbinterrupt, "on_kbinterrupt", ("raise", "return"))
        _assert_supported(on_blank, "on_blank", ("return", "pass"))
        _assert_supported(frontend, "frontend", ("simple", "fancy", "auto"))
        _assert_supported(return_mode, "return_mode", ("flat", "tree"))

        self.funcmap = construct_funcmap(cases, decorator=decorator)
        self.title = title
        self.once = once
        self.on_kbinterrupt = on_kbinterrupt
        self.case_args = case_args
        self.case_kwargs = case_kwargs
        self.case: Optional[str] = None  # Last registered case entered

        if self.case_args is None:
            self.case_args = {}
        if self.case_kwargs is None:
            self.case_kwargs = {}

        if on_blank == "return":
            self.on_blank = self._deactivate
        elif on_blank == "pass":
            self.on_blank = lambda: None

        # Special options
        self.special_cases = {"..": self.on_blank, "q": _quit, "h": print_help, "r": _restart}

        if frontend == "auto":
            self._frontend = _menu_simple
            try:
                import curses

                self._frontend = _menu_curses
            except ImportError:
                pass

        elif frontend == "fancy":
            self._frontend = _menu_curses
        elif frontend == "simple":
            self._frontend = _menu_simple

        if Menu._return_mode is None:
            Menu._return_mode = return_mode

        if Menu._return_mode == "flat":
            from meny.casehandlers import _FlatHandler

            self._case_handler: Callable = _FlatHandler()
        elif Menu._return_mode == "tree":
            from meny.casehandlers import _TreeHandler

            self._case_handler: Callable = _TreeHandler()

    def _deactivate(self):
        self.active = False

    def _menu_loop(self):
        """
        Menu loop
        """
        while self.active:
            inputstring: str = self._frontend(self)

            if cng.DEFAULT_CLEAR:
                clear_screen()

            if (not inputstring) or inputstring == "\n":
                self.on_blank()
                continue

            # Tokenize input
            try:
                inputlist: List[str] = input_splitter(inputstring)
            except ValueError as e:  # E.g. missing closing quotation or something
                _error_info_parse(e)
                continue

            # Get case
            self.case = inputlist.pop(0)

            # If self.case starts with '-', indicates reverse choice (like np.ndarray[-1])
            if self.case.startswith("-"):
                try:
                    self.case = str(len(self.funcmap) + int(self.case) + 1)
                except ValueError:
                    pass

            if self.case in self.funcmap:
                # Obtain self.case function from funcmap and
                # calls said function. Recall that items are
                # (description, function), hence the [1]
                casefunc = self.funcmap[self.case][1]
                self._case_handler(self, casefunc, inputlist)
            elif self.case in self.special_cases:
                # Items in special_cases are not tuples, but the
                # actual functions, so no need to do [1]
                casefunc = self.special_cases[self.case]
                self._case_handler(self, casefunc, inputlist)
            else:
                print(strings.INVALID_TERMINAL_INPUT_MSG)
                sleep(cng.MSG_WAIT_TIME)

            if self.once:
                self._deactivate()

    def run(self) -> Dict:
        """
        Responsibilities:
        - call menu loop
        - handle MenuQuit and KeyboardInterrupt
        - count depth
        """
        self.active = True
        Menu._depth += 1

        try:
            self._menu_loop()
        except KeyboardInterrupt:
            if self.on_kbinterrupt == "raise":
                self._deactivate()
                raise
            elif self.on_kbinterrupt == "return":
                print()
        except MenuQuit:
            if Menu._depth > 1:
                raise
        finally:
            Menu._depth -= 1
            if Menu._depth == 0:
                Menu._return_mode = None
                self._case_handler = None

        return Menu._return or {}


def build_menu(
    cases: Union[Iterable[FunctionType], Dict[str, FunctionType], ModuleType],
    title: Optional[str] = None,
    *,
    case_args: Optional[Dict[FunctionType, tuple]] = None,
    case_kwargs: Optional[Dict[FunctionType, dict]] = None,
    decorator: Optional[FunctionType] = None,
    frontend: Optional[str] = None,
    on_blank: Optional[str] = None,
    on_kbinterrupt: Optional[str] = None,
    once: Optional[bool] = None,
    return_mode: Optional[str] = None,
) -> Menu:
    """
    This is a factory for the Menu class to reduce boilerplate.
    See docstring in menu() below.
    Returns  Menu object
    """
    if isinstance(cases, ModuleType):
        cases_to_send = _get_module_cases(cases)
    elif isinstance(cases, dict):
        cases_to_send = _extract_and_preprocess_functions(cases)
        # If this menu is the first menu initialized, and is given the locally
        # defined functions, then must filter the functions that are defined
        # in __main__

        moduleName: Union[str, None] = cases.get("__name__", None)
        if moduleName == "__main__":
            cases_to_send = [case for case in cases_to_send if case.__module__ == "__main__"]

    elif isinstance(cases, Iterable):
        # Looks kinda stupid, but it reuses the code, which is nice
        cases_to_send = _extract_and_preprocess_functions({case.__name__: case for case in cases})
    else:
        raise TypeError(f"Invalid type for cases, got: {type(cases)}")

    cases_to_send: Iterable[FunctionType]

    cases_to_send = filter(lambda case: cng._CASE_IGNORE not in vars(case), cases_to_send)

    if once is None:
        once = cng.DEFAULT_ONCE
    return Menu(
        cases=cases_to_send,
        title=title or strings.DEFAULT_TITLE,
        case_args=case_args,
        case_kwargs=case_kwargs,
        decorator=decorator,
        frontend=frontend or cng.DEFAULT_FRONTEND,
        on_blank=on_blank or cng.DEFAULT_ON_BLANK,
        on_kbinterrupt=on_kbinterrupt or cng.DEFAULT_ON_INTERRUPT,
        once=once,
        return_mode=return_mode or cng.DEFAULT_RETURN_MODE,
    )


def menu(
    cases: Union[Iterable[FunctionType], Dict[str, FunctionType], ModuleType],
    title: Optional[str] = None,
    *,
    case_args: Optional[Dict[FunctionType, tuple]] = None,
    case_kwargs: Optional[Dict[FunctionType, dict]] = None,
    decorator: Optional[FunctionType] = None,
    frontend: Optional[str] = None,
    on_blank: Optional[str] = None,
    on_kbinterrupt: Optional[str] = None,
    once: Optional[bool] = None,
    return_mode: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Factory function for the CLI class. This function initializes a menu.

    ## Parameters
    - `cases`: can be
        - `Dict[str, FunctionType]`: a dictionary where keys are functions names and values are functions
        - `Iterable[FunctionType]` an iterable of functions
        - `ModuleType`: a module containing functions

    - `title`: title of menu

    - `cases_args`: dictionary with function as key and tuple of positional arguments as values

    - `cases_kwargs`: dictionary with function as key and dict of keyword arguments as values

    - `once`: If you want menu to return after a a single choice.

    - `on_blank`: What to do the when given blank input. Available options are:
        - `"return"`, will return to parent menu
        - `"pass"`, does nothing. This should only be used for the root menu.

    - `on_kbinterrupt`: Behavior when encountering KeyboardInterrupt exception when the menu is running.
                      If `"raise"`, then will raise `KeyboardInterrupt`, if `"return"` the menu returns.

    - `decorator`: Decorator to applied for all case functions.

    - `frontend`: specify desired frontend:
        - `"auto"`: Will try to use fancy frontend if curses module is available, else
                use simple frontend (default)
        - `"fancy"`: Use fancy front end (if on Windows, install
                    windows-curses first or Python will not be able to find the required
                    `"curses"` package that the fancy frontend uses)
        - `"simple"`: Use the simple (but compatible with basically everything) frontend

    - `return_mode`: the dictionary structure to be returned after the menu is done running. Only effective
        menu is root menu, as nested menus will use root's. Return mode options are:
        - `"flat"`: This is the default. Returns dictionary with function names (as `str`)
                as keys, and their return values as values (if they have been called), if not their names
                will not be in the dictionary (see examples). The downside of this return mode is if you have
                nested menus, where the nested menus reuse function names in from parent menus. The
                parent menus may overwrite the return values from the nested menus.
        - `"tree"`: Returns a nested dictionary structure, representing the structure of nested menus
                  (if you have that).

    ## Returns
    `Dict[str, Any]`: Dictionary where functions names are keys, and values are anything. Represents return
    values of case functions.

    ## Examples
    >>> def returnsOne():
    ...     def returnsOne():
    ...         return "1"
    ...     menu(locals())
    ...     return 1
    ...
    >>> def returnsTwo():
    ...     return 2
    ...
    >>> returns = menu(locals(), return_mode="flat") # Assume we have entered all cases
    { "returnsOne": 1, "returnsTwo": 2, }
    >>> returns = menu(locals(), return_mode="tree") # Assume we have entered all cases
    {
        "returnsOne": {
            "returnsOne": {
                "return": "1"
            }
            "return": 1
        },
        "returnsTwo": {
            "returns": 2
        },
    }
    >>> returns["returnsOne"]["returnsOne"]["return"]
    '1'
    """
    cli = build_menu(**locals())
    return cli.run()


if __name__ == "__main__":
    import subprocess

    subprocess.call(["python3", "example/cases.py"])



import curses
import curses.ascii
import inspect
from functools import wraps
from typing import Optional, Union

import meny
from meny import config as cng


def recover_cursor(f):
    """Wrapper for functions that should put cursor to where it was before"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        self: BaseWindow = args[0]
        prev_y, prev_x = self._window.getyx()
        return_val = f(*args, **kwargs)
        self._window.move(prev_y, prev_x)
        return return_val

    return wrapper


class BaseWindow:
    def __init__(self, window: "curses._CursesWindow"):
        self._window = window
        self._window.keypad(True)

    def cprint(self, string: str, *args, newline: int = 1):
        """
        Mimics print with newline
        """
        self._window.addstr(string, *args)
        y, x = self._window.getyx()
        if newline:
            self._window.move(y + newline, 0)


class InputField(BaseWindow):
    def __init__(self, main: "MainWindow") -> None:
        super().__init__(window=main._window)
        self.main: "MainWindow" = main
        self._inp: str = ""
        self.inp_message: str = "Input: "
        self.cprint(self.inp_message, newline=0)
        self.begin_y, self.begin_x = self._window.getyx()
        self.first_token: str = ""

    @property
    def inp(self):
        return self._inp

    @inp.setter
    def inp(self, string: str):
        self._inp = string
        self.first_token = self._inp.split(" ")[0]

    @recover_cursor
    def read_field(self, n: Optional[int] = None, decode: str = "utf-8") -> str:
        """
        Gets string in input field
        """
        if n:
            bytes_ = self._window.instr(self.begin_y, self.begin_x, n)
        else:
            bytes_ = self._window.instr(self.begin_y, self.begin_x)
        return bytes_.decode(decode).strip()

    def sync_window_with_inp(self):
        """
        Sets the visibe window to match the content of self._inp
        """
        self._window.move(self.begin_y, self.begin_x)
        self._window.clrtoeol()
        self._window.addstr(self.inp)

    def handle_backspace(self, y: int, x: int):
        if self.inp and x > self.begin_x:
            self._window.move(y, x - 1)
            self._window.delch()
            self.inp = self.read_field()

    def handle_int_input(self, k: int, y: int, x: int):
        """
        k: key
        y: current y pos
        x: current x pos
        """
        if k == curses.KEY_UP:
            if self.main.curr_index is None:
                self.inp = self.main.index2key[-1]
            else:
                new_index = (self.main.curr_index - 1) % len(self.main.funcmap)
                self.inp = self.main.index2key[new_index]
            self.inp += " "
            self.sync_window_with_inp()

        elif k == curses.KEY_DOWN:
            if self.main.curr_index is None:
                self.inp = self.main.index2key[0]
            else:
                new_index = (self.main.curr_index + 1) % len(self.main.funcmap)
                self.inp = self.main.index2key[new_index]
            self.inp += " "
            self.sync_window_with_inp()

        elif k == curses.KEY_RIGHT:
            _, max_x = self._window.getmaxyx()
            if (x < max_x) and (x < len(self.inp_message + self._inp)):
                self._window.move(y, x + 1)

        elif k == curses.KEY_LEFT:
            if x > len(self.inp_message):
                self._window.move(y, x - 1)

        elif k in (curses.KEY_BACKSPACE, curses.ascii.BS, curses.ascii.DEL):
            self.handle_backspace(y, x)

    def handle_str_input(self, k: str, y: int, x: int):
        if k == "\n":
            # Must capture newline explicitly, since insstr just treats it as space or something
            self.inp += "\n"
            self._window.clear()
        elif k in ("\b", "\x7f"):
            # Some systems (erm, Windows at least) gives "\b" for backspace
            self.handle_backspace(y, x)
        elif (k == "\x00") or (ord(k) == 0):
            # Windows key or some weird ass key, idk what to do about it, just return
            return
        else:
            # Sometimes crash on funny keys, e.g. windows key
            # use insstr since want to be able to insert new text in the middle
            # of existing text
            self._window.insstr(k)
            self._window.move(y, x + 1)
            self.inp = self.read_field()

    def handle_input(self, k: Union[int, str], y: int, x: int):
        if isinstance(k, str):
            self.handle_str_input(k, y, x)
        elif isinstance(y, int):
            self.handle_int_input(k, y, x)
        else:
            raise TypeError(f"k is of unexpected type: {type(k)}")


class MainWindow(BaseWindow):
    def __init__(self, cli: meny.Menu):
        self.cli = cli
        self.funcmap = cli.funcmap
        self.title = cli.title
        self.funcs_w_programmatic_args: set = cli.case_args.keys() | cli.case_kwargs.keys()
        self.available_cases = self.funcmap.keys() | cli.special_cases.keys()

        self.key2index = {key: index for index, key in enumerate(self.funcmap)}
        self.index2key = tuple(self.funcmap)
        self.curr_index: Optional[int] = None

    @property
    def prev_case(self):
        return self.cli.case

    @recover_cursor
    def highlight_funcmap(self, token: str, strlen: int):
        """Highlight line in funcmap print

        Exploits the fact that funcmap starts at line 1, and ends at line 1 + len(funcmap)

        Parameters
        ----------
        token : str
        strlen : int, how many columns highlight should span
        """
        # Handle option -1, -2 etc.. works like list(...)[-1]
        if token.startswith("-") and len(token) >= 2:
            try:
                token = str(len(self.funcmap) + int(token) + 1)
            except ValueError:
                pass

        new_index = self.key2index.get(token, None)

        if new_index is None:  # Index not in funcmap
            if self.curr_index is not None:
                # Clear previous highlight
                self._window.chgat(self.curr_index + 1, 0, curses.A_NORMAL)
                self.curr_index = None
            return

        if self.curr_index is not None:
            # If previously highlighted, remove highlight before highlighting new one
            self._window.chgat(self.curr_index + 1, 0, curses.A_NORMAL)

        self.curr_index = new_index
        self._window.chgat(new_index + 1, 0, strlen, curses.A_STANDOUT)

    @recover_cursor
    def hint_args(self, inp: str):
        prev_y, prev_x = self._window.getyx()
        # Clear line under input field
        self._window.move(prev_y + 1, 0)
        self._window.clrtoeol()
        # highlight_funcmap function sets self.curr_index
        if self.curr_index is None:
            return

        func = self.funcmap[self.index2key[self.curr_index]][1]
        signature = inspect.signature(func)
        paramiter = signature.parameters.items()
        iterlen = len(paramiter)
        # Dont show arghints for functions with programmatic arguments, and return if function
        # does not take any parameters
        if (iterlen == 0) or (func in self.funcs_w_programmatic_args):
            return

        self._window.move(prev_y + 1, prev_x)  # Under input field

        # Needs to split smartly to handle for quotations for string arguments
        input_split_error_flag: bool = False
        try:
            # Raises ValueError for malformed input caused by missing closing quotations
            inp_list = meny.input_splitter(inp)
        except ValueError as e:
            assert e.args[0] == "No closing quotation", (
                "shlex internals has changed, expected ValueError with args[0] == "
                f'"No closing quotation", but got "{type(e)}" with args: {e.args}\n'
                "Problem needs to be inspected by maintainer(s) of this package!"
            )
            inp_list = inp.split()
            input_split_error_flag = True

        n_tokens = len(inp_list)

        # Print argument hints, underline the current argument
        self._window.addstr("(")
        for i, p in enumerate(paramiter):
            if i == n_tokens - 2:
                self._window.addstr(f"{p[1]}", curses.A_UNDERLINE | curses.A_BOLD)
            else:
                self._window.addstr(f"{p[1]}")
            if i < iterlen - 1:
                self._window.addstr(", ")
        self._window.addstr(")")

        # Highlight with red if given too many arguments or if input parser complained
        if (n_tokens > (iterlen + 1)) or input_split_error_flag:
            self._window.chgat(prev_y + 1, prev_x, len(str(signature)), curses.color_pair(1))

    @recover_cursor
    def notify_special_token(self, inp: str, target_token: str, action: str):
        tokens = inp.strip().split(" ")
        if tokens[0] != target_token:
            return
        y, x = self._window.getyx()
        # Clear line under input field
        self._window.move(y + 1, 0)
        self._window.clrtoeol()

        self._window.move(y + 1, x)  # Under input field
        if len(tokens) == 1:
            message = f"Press enter to {action}"
            pair = curses.color_pair(2)
        else:
            message = f"Special case '{target_token}' does not accept arguments"
            pair = curses.color_pair(1)
        self._window.addstr(message)
        self._window.chgat(y + 1, x, len(message), pair)

    @recover_cursor
    def notify_invalid_case(self, first_token: str):
        if first_token in self.available_cases or not first_token:
            return

        y, x = self._window.getyx()
        message = "Invalid choice"
        self._window.move(y + 1, x)  # Under input field
        self._window.addstr("Invalid choice")
        self._window.chgat(y + 1, x, len(message), curses.color_pair(1))

    def run(self, window: "curses._CursesWindow"):
        """
        Will do almost all work on a padded window, which
        is accessible with self._window

        window (not self._window) is only used for keystrokes
        """
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)
        curses.init_pair(2, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        super().__init__(curses.newpad(2048, 2048))  # This will populate the self._window attribute

        def refresh_pad():
            lines, cols = window.getmaxyx()
            self._window.refresh(0, 0, 0, 0, lines - 1, cols - 1)

        # Print funcmap
        funcmap_strings = [f"{key}. {val[0]}" for key, val in self.funcmap.items()]
        maxstrlen = max(map(len, funcmap_strings))  # Get length of longest string in funcmap
        maxstrlen = max(maxstrlen, len(self.title))
        self.cprint(f"{self.title:^{maxstrlen}}", curses.A_UNDERLINE)
        for s in funcmap_strings:
            self.cprint(s)

        self.cprint("")

        inputfield = InputField(self)
        if cng.DEFAULT_REMEMBER and (self.prev_case is not None):
            inputfield.inp = self.prev_case + " "
            inputfield.sync_window_with_inp()
            self.highlight_funcmap(inputfield.first_token, maxstrlen)

        refresh_pad()
        while not inputfield._inp.endswith("\n"):
            # Get input using window instead of pad, using pad gives unexpected output
            k = self._window.get_wch()
            y, x = self._window.getyx()
            inputfield.handle_input(k, y, x)
            self.highlight_funcmap(inputfield.first_token, maxstrlen)
            self.hint_args(inputfield.inp)
            self.notify_special_token(inputfield.inp, "r", "restart")
            self.notify_special_token(inputfield.inp, "q", "quit menu(s)")
            self.notify_special_token(inputfield.inp, "h", "display help")
            self.notify_special_token(inputfield.inp, "..", "go back")
            self.notify_invalid_case(inputfield.first_token)
            refresh_pad()
        return inputfield._inp


def interface(cli: meny.Menu):
    _interface_object = MainWindow(cli)
    return curses.wrapper(_interface_object.run)



PURPLE = "\x1b[95m"
CYAN = "\x1b[96m"
DARKCYAN = "\x1b[36m"
BLUE = "\x1b[94m"
GREEN = "\x1b[92m"
YELLOW = "\x1b[93m"
RED = "\x1b[91m"
BOLD = "\x1b[1m"
UNDERLINE = "\x1b[4m"
END = "\x1b[0m"

START_MSG = "Starting console interface"
LOGO_TITLE = " Main menu "

ENTER_PROMPT = "Input"

EXIT_MSG = "Exiting program"
INVALID_TERMINAL_INPUT_MSG = "Invalid input"

INPUT_BLANK_HINT_RETURN = "Enter blank to return to parent menu"
INPUT_BLANK_HINT_PASS = ""
INPUT_BLANK_HINT_EXIT = "Enter blank to exit program"

INPUT_WAIT_PROMPT_MSG = "Press enter to continue"

DEFAULT_TITLE = " Title "



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="meny",
    version="1.0.1",
    author="Naphat Amundsen",
    author_email="naphat@live.com",
    description="Simple and sexy console interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Napam/meny",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "meny=meny.cli:cli"
        ]
    },
    python_requires=">=3.7",
)



from types import FunctionType

from meny.config import _CASE_IGNORE, _CASE_TITLE


def title(title: str):
    """
    Sets case title
    """

    def _title_appender(func: FunctionType):
        vars(func)[_CASE_TITLE] = title
        return func

    if isinstance(title, str):
        return _title_appender
    else:
        raise ValueError(f"Got unsupported type: {type(title)}")


def ignore(func: FunctionType):
    """
    Will flag case to be filtered out from menu creation
    """
    vars(func)[_CASE_IGNORE] = None  # Will only check for membership
    return func


if __name__ == "__main__":

    @title("Catdog")
    def testFunc1() -> str:
        pass

    @title("Catdog1")
    def testFunc2() -> str:
        pass

    print(vars(testFunc1))
    print(vars(testFunc2))
